from enum import StrEnum


class APIModel(StrEnum):
    DEFAULT = "gemini/gemini-2.0-flash-001"
    # DEFAULT = "claude-3-5-sonnet-20240620"
    OPENAI = "gpt-4o-2024-08-06"  # Default OpenAI model
    GPT_O3_MINI = "o3-mini-2025-01-31"
    GPT_4O_MINI = "gpt-4o-mini-2024-07-18"
    GPT_4O = "gpt-4o-2024-08-06"
    GEMINI = "gemini/gemini-2.0-flash-exp"  # Default Gemini model
    GEMINI_FLASH = "gemini/gemini-1.5-flash-002"
    GEMINI_PRO = "gemini/gemini-1.5-pro-002"
    GEMINI_FLASH_2 = "gemini/gemini-2.0-flash-001"
    GEMINI_FLASH_2_LITE = "gemini/gemini-2.0-flash-lite-preview-02-05"
    ANTHROPIC = "claude-3-5-sonnet-20240620"  # Default Anthropic model
    CLAUDE_SONNET = "claude-3-5-sonnet-20240620"
    CLAUDE_OPUS = "claude-3-opus-20240229"
    CLAUDE_HAIKU = "claude-3-5-haiku-20241022"


API_PROVIDERS = {
    APIModel.OPENAI: "OPENAI",
    APIModel.GPT_4O_MINI: "OPENAI",
    APIModel.GPT_4O: "OPENAI",
    APIModel.GEMINI: "GEMINI",
    APIModel.GEMINI_FLASH: "GEMINI",
    APIModel.GEMINI_PRO: "GEMINI",
    APIModel.GEMINI_FLASH_2: "GEMINI",
    APIModel.ANTHROPIC: "ANTHROPIC",
    APIModel.CLAUDE_SONNET: "ANTHROPIC",
    APIModel.CLAUDE_OPUS: "ANTHROPIC",
    APIModel.CLAUDE_HAIKU: "ANTHROPIC",
}



import pytest
from pydantic import BaseModel

from otaro.rule_utils import eval_rule_str


class SampleClass(BaseModel):
    foo: str
    a_list: list


SAMPLE = SampleClass(foo="bar", a_list=[1, 2, 3])


def test_eval_rule_str_basic():
    rule_eq_int = "otaro.rules.eq(1, 1)"
    assert eval_rule_str(rule_eq_int) is True
    rule_eq_int = "otaro.rules.eq('foo', 'foo')"
    assert eval_rule_str(rule_eq_int) is True


def eval_sample(rule_str: str):
    return eval_rule_str(
        rule_str=rule_str,
        sample=SAMPLE,
    )


def test_eval_rule_str_sample():
    # Basic equality
    assert eval_sample(f"otaro.rules.eq(foo, '{SAMPLE.foo}')") is True
    # Contains
    assert eval_sample(f"otaro.rules.contains(a_list, {SAMPLE.a_list[0]})") is True
    # Check length
    assert eval_sample(f"otaro.rules.length_eq(a_list, {len(SAMPLE.a_list)})") is True
    # Check list parsing
    assert eval_sample(f"otaro.rules.contains(['{SAMPLE.foo}'], foo)") is True
    assert eval_sample(f"otaro.rules.contains(['{SAMPLE.foo}',], foo)") is True
    assert (
        eval_sample(f"otaro.rules.contains(['{SAMPLE.foo}','{SAMPLE.foo}'], foo)")
        is True
    )
    # Check tuple parsing
    assert eval_sample(f"otaro.rules.contains(('{SAMPLE.foo}'), foo)") is True
    assert eval_sample(f"otaro.rules.contains(('{SAMPLE.foo}',), foo)") is True
    assert (
        eval_sample(f"otaro.rules.contains(('{SAMPLE.foo}','{SAMPLE.foo}'), foo)")
        is True
    )


def test_eval_rule_str_kwargs():
    # Test kwargs
    assert eval_sample(f"otaro.rules.eq(foo, b='{SAMPLE.foo}')")
    assert eval_sample(f"otaro.rules.eq(a=foo, b='{SAMPLE.foo}')")


def test_eval_rule_str_no_args():
    # Test rules with no args
    assert eval_sample("otaro.rules.always_true") is True
    assert eval_sample("otaro.rules.always_false") is False


def test_invalid_rule():
    # Check that error is raised correctly
    invalid_rule = "invalid rule"
    with pytest.raises(ValueError, match="Invalid rule"):
        eval_rule_str(invalid_rule)



import logging
from pathlib import Path

from otaro.task import Task

logging.basicConfig()
logger = logging.getLogger("otaro.tests.test_task_run")
logger.setLevel(logging.INFO)


def test_task_run():
    task = Task.from_config("tests/math.yaml")
    response = task.run(equation="40 + 2")
    assert response.answer == 42


def test_task_optim():
    task = Task.from_config("tests/poet.yaml")
    response = task.run(topic="life")
    assert "frog" in response.haiku.lower()
    optimized_file = Path("./tests/poet.optim.yaml")
    assert optimized_file.exists() is True
    optimized_desc = task.desc
    # Check that optimization loads by default
    new_task = Task.from_config("tests/poet.yaml")
    assert new_task.desc == optimized_desc
    optimized_file.unlink(missing_ok=True)



import os
from pathlib import Path

from otaro.task import Task


def main():
    file_dir = Path(os.path.dirname(os.path.realpath(__file__)))
    task = Task.from_config(file_dir / "planner.yaml")
    response = task.run(
        task="Search Google for news on AI and save it to a CSV", optimize=False
    )
    print("Plan:")
    for step in response.plan:
        print(f"{step.number} [{step.status.value}]: {step.desc}")


if __name__ == "__main__":
    main()



from pydantic import BaseModel


# For testing
def always_true(_: BaseModel):
    return True


# For testing
def always_false(_: BaseModel):
    return False


def eq(a, b):
    return a == b


def neq(a, b):
    return a != b


def contains(a, b):
    if isinstance(a, str):
        a = a.lower()
    if isinstance(b, str):
        b = b.lower()
    return b in a


def length_eq(a, b: int):
    return len(a) == b



import logging

from otaro.task import Task

logging.basicConfig()
logger = logging.getLogger("otaro.tests.test_task_load")
logger.setLevel(logging.INFO)


with open("./tests/math_system_prompt.txt", "r") as file:
    MATH_SYSTEM_PROMPT = file.read()

with open("./tests/nested_system_prompt.txt", "r") as file:
    NESTED_SYSTEM_PROMPT = file.read()

with open("./tests/nested_user_prompt.txt", "r") as file:
    NESTED_USER_PROMPT = file.read()


def test_load_config():
    task = Task.from_config("tests/math.yaml")
    assert task.desc == ""
    assert len(task.inputs) == 1
    assert len(task.outputs) == 1
    prompt_template = task.prompt_template
    assert prompt_template["messages"][0]["content"] == MATH_SYSTEM_PROMPT


def test_load_nested():
    task = Task.from_config("tests/nested.yaml")
    system_prompt = task.prompt_template["messages"][0]["content"]
    # logger.info(system_prompt)
    dummy_input = task.dummy_input
    # logger.info(dummy_input)
    user_prompt = task.get_prompt(**dummy_input)["messages"][1]["content"]
    # logger.info(user_prompt)
    assert system_prompt == NESTED_SYSTEM_PROMPT
    assert user_prompt == NESTED_USER_PROMPT

    for input_field in task.inputs:
        input_field.parse(str(input_field.dummy_value))



from otaro.types import Field


def test_simple_field_attributes():
    field = Field(
        name="foo",
        type="str",
        desc="Hello, world",
    )
    assert field.model == str
    assert field.dummy_value == "Foo"
    assert str(field) == "`foo` (str): Hello, world."



import asyncio
import logging
import re
from pathlib import Path

import yaml
from litellm import acompletion, token_counter
from pydantic import BaseModel, create_model
from pydantic.fields import FieldInfo
from tqdm import tqdm

from otaro.prompts import (
    SYSTEM_PROMPT,
    USER_PROMPT,
)
from otaro.providers import APIModel
from otaro.rule_utils import eval_rule_str, get_rule_source
from otaro.task_utils import parse_fields_config
from otaro.types import Field

logging.basicConfig()
logger = logging.getLogger("otaro.task")
logger.setLevel(logging.INFO)


class CompletionResponse(BaseModel):
    content: str
    num_input_tokens: int
    num_output_tokens: int


def count_tokens(
    message: str,
    role="user",
    model: str | None = None,
):
    return token_counter(
        model=model or APIModel.DEFAULT,
        messages=[
            {
                "role": role,
                "content": message,
            }
        ],
    )


async def _completion(
    model: str,
    messages: list[dict],
):
    model = model
    response = await acompletion(
        model=model,
        messages=messages,
        num_retries=3,
        timeout=60,
    )
    content = response.choices[0].message.content
    num_input_tokens = token_counter(
        model=model,
        messages=messages,
    )
    num_output_tokens = token_counter(
        model=model,
        messages=[
            {
                "role": "assistant",
                "content": content,
            }
        ],
    )
    return CompletionResponse(
        content=content,
        num_input_tokens=num_input_tokens,
        num_output_tokens=num_output_tokens,
    )


class Task(BaseModel):
    model: str
    inputs: list[Field]
    outputs: list[Field]
    desc: str = ""
    demos: list[dict] = []
    rules: list[str] = []
    config_file: str | None = None

    def __init__(
        self,
        model: str,
        inputs: list[Field],
        outputs: list[Field],
        desc: str = "",
        demos: list[dict] | None = None,
        rules: list[str] | None = None,
        config_file: str | None = None,
        **kwargs,
    ):
        super().__init__(
            model=model,
            inputs=inputs or [],
            outputs=outputs or [],
            desc=desc,
            demos=demos or [],
            rules=rules or [],
            config_file=config_file,
        )

    # TODO: Check for duplicate field name between inputs and outputs
    @classmethod
    def from_config(cls, config_file: str | Path):
        config_file = Path(config_file)
        with open(config_file) as file:
            config = yaml.safe_load(file)
        config["inputs"] = parse_fields_config(config.get("inputs", {}))
        config["outputs"] = parse_fields_config(config.get("outputs", {}))
        # rules = []
        # module = importlib.import_module("otaro.rules")
        # for rule in config.get("rules", []):
        #     print(rule)
        #     fn_pattern = re.compile(
        #         "^(?P<module>.*?)\\.(?P<fn>.*?)(\\((?P<args>.*)\\))?$", re.DOTALL
        #     )
        #     match = fn_pattern.match(rule.strip())
        #     if match:
        #         print(match.groupdict())
        #         args_str = match.groupdict().get("args")
        #         if args_str:
        #             args = process_signature(args_str)
        #             print(args)
        #         quit()
        #         module_name = match.groupdict().get("module")
        #         fn_name = match.groupdict().get("fn")
        #         module = importlib.import_module(module_name)
        #         fn = getattr(module, fn_name)
        #         rules.append(fn)
        #     quit()
        # config["rules"] = rules

        optimized_config_file = config_file.parent / (config_file.stem + ".optim.yaml")
        optimized_config = {}
        if optimized_config_file.exists():
            with open(optimized_config_file) as file:
                optimized_config = yaml.safe_load(file)
            # optimized_config = cls.from_config(optimized_config_file).model_dump(
            #     mode="json"
            # )
            # del optimized_config["config_file"]
            # for attr in list(optimized_config.keys()):
            #     if not optimized_config[attr]:
            #         del optimized_config[attr]
        config.update(optimized_config)
        # logger.info(f"Config: {config}")
        return cls(config_file=str(config_file), **config)

    @property
    def formatted_desc(self):
        default_desc = f"Given the field{'s' if len(self.inputs) > 1 else ''} {', '.join(f.name for f in self.inputs)}, produce the field{'s' if len(self.outputs) > 1 else ''} {', '.join(f.name for f in self.outputs)}"
        if self.desc:
            return f"{self.desc} {default_desc}"
        else:
            return default_desc

    @property
    def dummy_input(self):
        dummy_inputs = {}
        for input_field in self.inputs:
            dummy_inputs[input_field.name] = input_field.dummy_value
        return dummy_inputs

    @property
    def prompt_template(self):
        input_fields = "\n".join(
            [f"{i}. {field}" for i, field in enumerate(self.inputs, 1)]
        )
        output_fields = "\n".join(
            ["1. `reasoning` (str)"]
            + [f"{i}. {field}" for i, field in enumerate(self.outputs, 2)]
        )
        interaction_format = "\n\n".join(
            [field.dummy_template for field in self.inputs]
            + ["[[ ## reasoning ## ]]\n{reasoning}"]
            + [field.dummy_template for field in self.outputs]
        )
        system_prompt = SYSTEM_PROMPT.format(
            input_fields=input_fields,
            output_fields=output_fields,
            interaction_format=interaction_format,
            objective=self.formatted_desc,
        )

        user_prompt = USER_PROMPT.format(
            input_values="\n\n".join(field.dummy_template for field in self.inputs),
            output_fields=", ".join(["reasoning"] + [f.name for f in self.outputs]),
        )

        messages = [
            {"role": "system", "content": system_prompt},
        ]

        for demo_kwargs in self.demos:
            demo_user_prompt = USER_PROMPT.format(
                input_values="\n\n".join(
                    field.template(demo_kwargs[field.name]) for field in self.inputs
                ),
                output_fields=", ".join(["reasoning"] + [f.name for f in self.outputs]),
            )
            demo_assistant_prompt = (
                "\n\n".join(
                    field.template(demo_kwargs[field.name]) for field in self.outputs
                )
                + "\n\n[[ ## completed ## ]]\n"
            )
            if "reasoning" in demo_kwargs:
                demo_assistant_prompt = (
                    f"[[ ## reasoning ## ]]\n{demo_kwargs['reasoning']}\n\n"
                    + demo_assistant_prompt
                )
            messages.append({"role": "user", "content": demo_user_prompt})
            messages.append({"role": "assistant", "content": demo_assistant_prompt})

        messages.append({"role": "user", "content": user_prompt})

        return {"messages": messages}

    def get_prompt(self, **kwargs):
        input_fields = "\n".join(
            [f"{i}. {field}" for i, field in enumerate(self.inputs, 1)]
        )
        output_fields = "\n".join(
            ["1. `reasoning` (str)"]
            + [f"{i}. {field}" for i, field in enumerate(self.outputs, 2)]
        )
        interaction_format = "\n\n".join(
            [field.dummy_template for field in self.inputs]
            + ["[[ ## reasoning ## ]]\n{reasoning}"]
            + [field.dummy_template for field in self.outputs]
        )
        system_prompt = SYSTEM_PROMPT.format(
            input_fields=input_fields,
            output_fields=output_fields,
            interaction_format=interaction_format,
            objective=self.formatted_desc,
        )

        user_prompt = USER_PROMPT.format(
            input_values="\n\n".join(
                field.template(kwargs[field.name]) for field in self.inputs
            ),
            output_fields=", ".join(["reasoning"] + [f.name for f in self.outputs]),
        )

        messages = [
            {"role": "system", "content": system_prompt},
        ]

        for demo_kwargs in self.demos:
            demo_user_prompt = USER_PROMPT.format(
                input_values="\n\n".join(
                    field.template(demo_kwargs[field.name]) for field in self.inputs
                ),
                output_fields=", ".join(["reasoning"] + [f.name for f in self.outputs]),
            )
            demo_assistant_prompt = (
                "\n\n".join(
                    field.template(demo_kwargs[field.name]) for field in self.outputs
                )
                + "\n\n[[ ## completed ## ]]\n"
            )
            if "reasoning" in demo_kwargs:
                demo_assistant_prompt = (
                    f"[[ ## reasoning ## ]]\n{demo_kwargs['reasoning']}\n\n"
                    + demo_assistant_prompt
                )
            messages.append({"role": "user", "content": demo_user_prompt})
            messages.append({"role": "assistant", "content": demo_assistant_prompt})

        messages.append({"role": "user", "content": user_prompt})

        return {"messages": messages}

    def run(self, optimize=True, **kwargs):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self.arun(optimize, **kwargs))

    async def arun(self, optimize=True, **kwargs):
        messages = self.get_prompt(**kwargs).get("messages")
        result = None

        num_tries = 3

        for _ in range(num_tries):
            response = await _completion(model=self.model, messages=messages)

            output_field_names = ["reasoning"] + [field.name for field in self.outputs]
            rgx_patterns = [
                f"\\[\\[ #?#? ?{field_name} #?#? ?\\]\\](?P<{field_name}>.*?)"
                for field_name in output_field_names
            ]
            rgx_pattern = (
                ".*?" + "".join(rgx_patterns) + "\\[\\[ #?#? ?completed #?#? ?\\]\\]"
            )

            rgx = re.compile(rgx_pattern, re.DOTALL)

            match = rgx.match(response.content)
            if match is None:
                logger.warning(f"No match on {response.content}")
            else:
                try:
                    input_field_attributes = {}
                    for field in self.inputs:
                        input_field_attributes[field.name] = (field.model, FieldInfo())
                    output_field_attributes = {}
                    for field in self.outputs:
                        output_field_attributes[field.name] = (field.model, FieldInfo())
                    model = create_model(
                        "Output",
                        reasoning=(str, FieldInfo()),
                        **output_field_attributes,
                        **input_field_attributes,
                        num_input_tokens=(int, FieldInfo()),
                        num_output_tokens=(int, FieldInfo()),
                    )
                    groupdict = {
                        "num_input_tokens": response.num_input_tokens,
                        "num_output_tokens": response.num_output_tokens,
                    }
                    for k, v in match.groupdict().items():
                        if k == "reasoning":
                            groupdict[k] = str(v.strip())
                        else:
                            output_field = [o for o in self.outputs if o.name == k][0]
                            groupdict[k] = output_field.parse(v.strip(), to_dict=True)
                    result = model(**groupdict, **kwargs)
                except Exception as e:
                    result = str(e)
                    logger.exception(e)

                # Optimize against rules
                evals = [False] * len(self.rules)
                if len(self.rules):
                    logger.info("Evaluating rules")
                    for i, rule in enumerate(self.rules, 1):
                        outcome = eval_rule_str(rule, result)
                        logger.info(f"Rule {i}: {rule} - {outcome}")
                        evals[i - 1] = outcome

                if optimize and self.config_file and len(self.rules):
                    if not all(evals):
                        new_task = await self.optimize([kwargs], num_prompts=1)
                        # TODO: Export more optimized params
                        config_file = Path(self.config_file)
                        optimized_config_file = config_file.parent / (
                            config_file.stem + ".optim.yaml"
                        )
                        with open(optimized_config_file, "w") as file:
                            yaml.safe_dump(
                                {
                                    "desc": new_task.desc,
                                },
                                file,
                                default_flow_style=False,
                            )
                        result = await new_task.arun(**kwargs, optimize=False)
                        # Update to optimized task
                        self.sync_to(new_task)

                return result

        raise ValueError("Unable to generate valid response")

    async def evaluate(self, data: list):
        total_score = 0
        num_evals = 0
        for sample in tqdm(data):
            try:
                result = await self.arun(**sample, optimize=False)
                rule_evals = []
                for rule in self.rules:
                    outcome = eval_rule_str(rule, result)
                    rule_evals.append(outcome)
                if all(rule_evals):
                    total_score += 1.0
            except Exception as e:
                logger.exception(e)
            num_evals += 1.0
        return total_score / num_evals

    async def optimize(self, data: list, say=None, num_prompts=5):
        def _log(message: str):
            if say:
                say(message)
            logger.info(message)

        _log("Optimizing task...")

        # Run initial evaluation with rules to determine success rate
        initial_score = await self.evaluate(data)
        _log(f"Initial score: {initial_score}")

        # Get prompts
        prompts = await get_prompts(self, num_prompts=num_prompts)

        # For each prompt, run evaluation
        # TODO: Add prompt-demo combination
        scores = []
        for prompt in prompts:
            _log(f"Evaluating prompt: {prompt}")
            task_variant = self.model_copy(deep=True)
            task_variant.desc = prompt
            score = await task_variant.evaluate(data)
            scores.append(score)
            _log(f"\tScore: {score}")
            _log(f"All scores: {[initial_score] + scores}")

        # Select best parameters
        if max(scores) >= initial_score:
            all_scores = [initial_score] + scores
            all_prompts = [self.desc] + prompts

            # Select best-performing prompt that is the shortest
            best_idxs = [
                i for i, score in enumerate(all_scores) if score >= max(all_scores)
            ]
            best_prompts = [all_prompts[i] for i in best_idxs]
            prompt_lengths = [count_tokens(p) for p in best_prompts]
            shortest_idxs = [
                i
                for i, length in enumerate(prompt_lengths)
                if length <= min(prompt_lengths)
            ]
            best_prompt = best_prompts[shortest_idxs[0]]
            best_idx = all_prompts.index(best_prompt)
            _log(f"Selecting prompt #{best_idx} with score {all_scores[best_idx]}")
            task_variant = self.model_copy(deep=True)
            task_variant.desc = best_prompt
            return task_variant
        else:
            _log("Optimization failed")
            return self

    def sync_to(self, new_task: "Task"):
        # Used to update self to optimized version
        # TODO: Sync more attributes
        self.desc = new_task.desc


# TODO: Suggest prompts based on successful and unsuccessful examples
async def get_prompts(
    task: Task, model: str | None = None, num_prompts: int = 5
) -> list[str]:
    optim_task = Task(
        model=model or task.model,
        desc=(
            "An LLM needs to be prompted to accomplish the described task. "
            "The LLM's output has to fulfill all of the rules provided. "
            "Suggest prompts in natural language for this LLM to maximize its chances of success. "
            "Each prompt will be shown after all of the input/output schemas are displayed. "
            "Use the correct input/output names in the prompts where relevant. "
            "Do NOT include instructions about JSON schema formats. "
            "Minimize the number of words in the prompts. "
            "Highlight if the task cannot be completed due to contradicting rules or incomplete task descriptions. "
        ),
        inputs=[
            Field(
                name="task_description",
                type="str",
            ),
            Field(
                name="input_schemas",
                type="list",
                list_child_type=Field(
                    name="schema",
                    type="str",
                ),
            ),
            Field(
                name="output_schemas",
                type="list",
                list_child_type=Field(
                    name="schema",
                    type="str",
                ),
            ),
            Field(
                name="rules",
                type="list",
                list_child_type=Field(
                    name="rule",
                    type="str",
                ),
            ),
            Field(
                name="num_prompts",
                type="int",
            ),
        ],
        outputs=[
            Field(
                name="prompts",
                type="list",
                list_child_type=Field(
                    name="prompt",
                    type="str",
                ),
            )
        ],
    )

    max_tries = 3
    for _ in range(max_tries):
        kwargs = {
            "task_description": task.desc,
            "input_schemas": [str(i) for i in task.inputs],
            "output_schemas": [str(i) for i in task.outputs],
            # TODO: Need to handle source different for rules with and without args
            # Current getsource only works with rules that do not have args i.e. takes in sample
            # For rules with args e.g. otaro.rules.contains(haiku, "green") we probably need to
            # also pass the rule string
            "rules": [get_rule_source(i) for i in task.rules],
            "num_prompts": num_prompts,
        }
        result = await optim_task.arun(**kwargs, optimize=False)
        if not isinstance(result.prompts, str) and len(result.prompts) == num_prompts:
            break

    return result.prompts



from dotenv import load_dotenv
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

load_dotenv()


app = FastAPI()
origins = [
    "*",
    "http://0.0.0.0:8081",
    "http://localhost:8081",
    "http://127.0.0.1:8081",
]
app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


@app.get("/")
def get_status():
    return {
        "status": "running",
    }



import json
import logging
from enum import Enum
from typing import Union

from pydantic import BaseModel, create_model
from pydantic.fields import FieldInfo

from otaro.parsing import llm_parse_json

logging.basicConfig()
logger = logging.getLogger("otaro.types")
logger.setLevel(logging.INFO)


class FieldType(str, Enum):
    STR = "str"
    INT = "int"
    FLOAT = "float"
    BOOL = "bool"
    ENUM = "enum"
    LIST = "list"
    OBJECT = "object"


class Field(BaseModel):
    name: str
    type: FieldType
    desc: str = ""
    enum_members: Union[list[str], None] = None
    list_child_type: Union["Field", None] = None
    object_attributes: Union[list["Field"], None] = None

    def __init__(
        self,
        name: str,
        type: FieldType = FieldType.STR,
        desc: str = "",
        enum_members: Union[list[str], None] = None,
        list_child_type: Union["Field", None] = None,
        object_attributes: list["Field"] | None = None,
    ):
        super().__init__(
            name=name,
            type=type,
            desc=desc,
            enum_members=enum_members,
            list_child_type=list_child_type,
            object_attributes=object_attributes or [],
        )
        self.type = type
        self.enum_members = enum_members or []
        self.enum_members = [m.strip() for m in self.enum_members]
        if list_child_type:
            if isinstance(list_child_type, Field):
                self.list_child_type = list_child_type
            else:
                self.list_child_type = Field(**list_child_type)
        if object_attributes:
            self.object_attributes = [Field(**attr) for attr in object_attributes]

    @property
    def model(self):
        if self.type == FieldType.STR:
            return str
        elif self.type == FieldType.INT:
            return int
        elif self.type == FieldType.FLOAT:
            return float
        elif self.type == FieldType.BOOL:
            return bool
        elif self.type == FieldType.ENUM:
            return Enum(
                self.name,
                [(str(v), v) for v in self.enum_members],
                type=type(
                    self.enum_members[0]
                ),  # Mix type of first member (assuming all members are same type)
            )
        elif self.type == FieldType.LIST:
            return list[self.list_child_type.model]
        elif self.type == FieldType.OBJECT:
            attributes = {}
            for attr in self.object_attributes:
                attributes[attr.name] = (attr.model, FieldInfo())
            model = create_model(self.name, **attributes)
            return model

    @property
    def schema(self):
        if self.type in [
            FieldType.STR,
            FieldType.INT,
            FieldType.FLOAT,
            FieldType.BOOL,
        ]:
            return self.type
        elif self.type == FieldType.ENUM:
            return f'{{"enum": {json.dumps(self.enum_members)}}}'
        elif self.type == FieldType.LIST:
            # TODO: This should be recursive
            if self.list_child_type.type in [
                FieldType.STR,
                FieldType.INT,
                FieldType.FLOAT,
                FieldType.BOOL,
            ]:
                return f'{{"items": {self.list_child_type.type}, "type": "array"}}'
            if self.list_child_type.type == FieldType.ENUM:
                return f'{{"items": {self.list_child_type.schema}, "type": "array"}}'
            else:
                return f'{{"$defs": {{"{self.list_child_type.name.capitalize()}": {self.list_child_type.schema}}}, "items": {{"$ref": "#/$defs/{self.list_child_type.name.capitalize()}"}}, "type": "array"}}'
        elif self.type == FieldType.OBJECT:
            return self.model.model_json_schema()

    @property
    def dummy_value(self):
        if self.type == FieldType.STR:
            return "Foo"
        elif self.type == FieldType.INT:
            return 42
        elif self.type == FieldType.FLOAT:
            return 3.14
        elif self.type == FieldType.BOOL:
            return True
        elif self.type == FieldType.ENUM:
            return self.enum_members[0]
        elif self.type == FieldType.LIST:
            return [self.list_child_type.dummy_value]
        elif self.type == FieldType.OBJECT:
            attribute_values = {}
            for attr in self.object_attributes:
                attribute_values[attr.name] = attr.dummy_value
            return attribute_values

    def __str__(self):
        docstring = f"`{self.name}` ({self.type})"
        details = [
            self.desc + ("." if len(self.desc) and not self.desc.endswith(".") else "")
        ]
        if self.type == FieldType.ENUM:
            details.append("Respond with one of the following enum values.")
            details.append(f"Enum Schema: {self.schema}")
        if self.type in [FieldType.LIST, FieldType.OBJECT]:
            if self.type == FieldType.LIST:
                details.append("Respond with a single JSON array.")
            elif self.type == FieldType.OBJECT:
                details.append("Respond with a single JSON object.")
            details.append(f"JSON Schema: {self.schema}")
        details_str = " ".join(details)
        if len(details_str):
            docstring += f": {details_str}"
        return docstring

    @property
    def dummy_template(self):
        return f"[[ ## {self.name} ## ]]\n{{{self.name}}}"

    def template(self, value):
        if isinstance(value, (str, float, int)):
            return f"[[ ## {self.name} ## ]]\n{value}"
        else:
            return f"[[ ## {self.name} ## ]]\n{json.dumps(value)}"

    def parse(self, value: str, to_dict=False):
        if self.type == FieldType.STR:
            return value
        elif self.type == FieldType.INT:
            return int(value)
        elif self.type == FieldType.FLOAT:
            return float(value)
        elif self.type == FieldType.BOOL:
            if isinstance(value, str) and value.lower() in [
                "false",
                "f",
                "no",
                "0",
                "null",
                "none",
                "undefined",
            ]:
                return False
            return bool(value)
        elif self.type == FieldType.ENUM:
            parsed_value = self.model(value)
            if to_dict:
                return parsed_value.value
            else:
                return parsed_value
        elif self.type == FieldType.LIST:
            value_json = llm_parse_json(value)
            return [
                self.list_child_type.parse(json.dumps(c), to_dict=to_dict)
                for c in value_json
            ]
        elif self.type == FieldType.OBJECT:
            value_json = llm_parse_json(value)
            parsed_value = self.model(**value_json)
            if to_dict:
                return parsed_value.model_dump(mode="json")
            else:
                return parsed_value

    def __call__(self, value):
        return self.parse(value)


# OPERATORS = {
#     "<": lambda a, b: a < b,
#     ">": lambda a, b: a > b,
#     "<=": lambda a, b: a <= b,
#     ">=": lambda a, b: a >= b,
#     "!=": lambda a, b: a != b,
#     "==": lambda a, b: a == b,
#     "in": lambda a, b: a in b,
#     "not": lambda a, _: not a,
#     "and": lambda a, b: a and b,
#     "or": lambda a, b: a or b,
#     "len": lambda a, _: len(a),
# }


# class RuleOperand(BaseModel):
#     value: str
#     type: str

#     def __init__(
#         self,
#         type: str = "",
#         value: str = "",
#         **kwargs,
#     ):
#         super().__init__(
#             type=type,
#             value=value,
#         )
#         self.type = type

#     def __str__(self):
#         return f"{self.type}<{self.value}>"

#     def eval(self, result=None):
#         if self.type == "field":
#             props = self.value.split(".")
#             subresult = result.model_dump(mode="json")
#             for prop in props:
#                 match = re.match(re.compile("(?P<prop>.*?)\\[(?P<idx>\\d*)\\]"), prop)
#                 if match:
#                     prop = match.groupdict()["prop"]
#                     idx = int(match.groupdict()["idx"])
#                     subresult = subresult[prop][idx]
#                 else:
#                     subresult = subresult[prop]
#             return subresult
#         if self.type == "str":
#             return self.value
#         elif self.type == "float":
#             return float(self.value)
#         elif self.type == "int":
#             return int(self.value)


# class Rule(BaseModel):
#     a: Union[RuleOperand, "Rule"]
#     operator: str
#     b: Union[RuleOperand, "Rule", None] = None
#     type: str

#     def __init__(
#         self,
#         a: Union[RuleOperand, "Rule", None] = None,
#         operator: str = "",
#         b: Union[RuleOperand, "Rule", None] = None,
#         type: str = "",
#         **kwargs,
#     ):
#         super().__init__(
#             a=a,
#             operator=operator,
#             b=b,
#             type=type,
#         )
#         if self.a.type == "rule":
#             self.a = Rule(**a)
#         if self.b and self.b.type == "rule":
#             self.b = Rule(**b)

#     def eval(self, result=None):
#         parsed_a = self.a.eval(result)
#         if self.b:
#             parsed_b = self.b.eval(result)
#         else:
#             parsed_b = None
#         return OPERATORS[self.operator](parsed_a, parsed_b)

#     def __str__(self):
#         if self.operator in ["not", "len"]:
#             return f"{self.operator}({self.a})"
#         else:
#             return f"({self.a}) {self.operator} ({self.b})"


# class TaskParams(BaseModel):
#     desc: str = ""
#     inputs: list[Field] = []
#     outputs: list[Field] = []
#     rules: list[str] = []
#     data: list[dict] = []


# class OptimParams(BaseModel):
#     desc: str = ""
#     inputs: list[Field] = []
#     outputs: list[Field] = []
#     rules: list[str] = []
#     data: list[dict] = []



import logging
import re
from typing import Any

import dirtyjson
from dirtyjson.attributed_containers import AttributedDict, AttributedList

logging.basicConfig()
logger = logging.getLogger("otaro.parsing")
logger.setLevel(logging.INFO)


def find_json(rgx: str, text: str):
    match = re.search(rgx, text)
    if match is None:
        return text
    else:
        return match.groupdict().get("json")


def convert_attributed_container(
    container: Any | AttributedDict | AttributedList | float | int,
):
    if isinstance(container, AttributedList):
        return [convert_attributed_container(i) for i in container]
    elif isinstance(container, AttributedDict):
        dict_container = {**container}
        for k, v in dict_container.items():
            dict_container[k] = convert_attributed_container(v)
        return dict_container
    else:
        return container


def llm_parse_json(text: str):
    """Read LLM output and extract JSON data from it."""

    # First check for ```json
    code_snippet_pattern = r"```json(?P<json>(.|\s|\n)*?)```"
    code_snippet_result = find_json(code_snippet_pattern, text)
    # Then try to find the longer match between [.*?] and {.*?}
    array_pattern = re.compile("(?P<json>\\[.*\\])", re.DOTALL)
    array_result = find_json(array_pattern, text)
    dict_pattern = re.compile("(?P<json>{.*})", re.DOTALL)
    dict_result = find_json(dict_pattern, text)

    if array_result and dict_result and len(dict_result) > len(array_result):
        results = [
            code_snippet_result,
            dict_result,
            array_result,
        ]
    else:
        results = [
            code_snippet_result,
            array_result,
            dict_result,
        ]

    # Try each result in order
    for result in results:
        if result is not None:
            try:
                result_json = dirtyjson.loads(result)
                return convert_attributed_container(result_json)
            except dirtyjson.error.Error:
                pass
            try:
                result = (
                    result.replace("None", "null")
                    .replace("True", "true")
                    .replace("False", "false")
                )
                result_json = dirtyjson.loads(result)
                return convert_attributed_container(result_json)
            except dirtyjson.error.Error:
                continue

    error_message = (
        f"Failed to parse JSON from text {text!r}. Your model may not be capable of"
        " supporting JSON output or our parsing technique could use some work. Try"
        " a different model"
    )
    raise ValueError(error_message)



SYSTEM_PROMPT = """Your input fields are:
{input_fields}

Your output fields are:
{output_fields}

All interactions will be structured in the following way, with the appropriate values filled in.

{interaction_format}

[[ ## completed ## ]]

In adhering to this structure, your objective is:
    {objective}."""



from pydantic import BaseModel


def haiku_contains_green(sample: BaseModel):
    return "green" in sample.haiku.lower()


def haiku_contains_blue(sample: BaseModel):
    return "blue" in sample.haiku.lower()



import logging

logging.basicConfig()
logger = logging.getLogger("otaro.task_utils")
logger.setLevel(logging.INFO)


# TODO: Provide more detailed error messages
def parse_attr_config(attr_json: dict):
    if attr_json.get("type") == "enum":
        enum_members = attr_json.get("enum_members", [])
        if not len(enum_members):
            raise ValueError("No enum_members provided")
    if attr_json.get("type") == "list":
        list_child_type_config = attr_json.get("list_child_type")
        if list_child_type_config:
            keys = list(list_child_type_config.keys())
            if len(keys) != 1:
                raise ValueError("Invalid list_child_type")
            attr_json["list_child_type"] = {
                "name": keys[0],
                **parse_attr_config(list_child_type_config[keys[0]]),
            }
        else:
            raise ValueError("No list_child_type provided")
    if attr_json.get("type") == "object":
        object_attributes_config = attr_json.get("object_attributes")
        if object_attributes_config:
            attr_json["object_attributes"] = [
                {"name": name, **parse_attr_config(attrs)}
                for name, attrs in object_attributes_config.items()
            ]
        else:
            raise ValueError("No object_attributes provided")

    return attr_json


def parse_fields_config(json: dict):
    return [
        {
            "name": name,
            **parse_attr_config(attr if attr else {}),
        }
        for name, attr in json.items()
    ]



import os
from pathlib import Path

from otaro.task import Task


def main():
    file_dir = Path(os.path.dirname(os.path.realpath(__file__)))
    task = Task.from_config(file_dir / "poet.yaml")
    response = task.run(topic="lion", optimize=False)
    print(f"Haiku:\n{response.haiku}")


if __name__ == "__main__":
    main()



import json

import pytest

from otaro.parsing import llm_parse_json

SAMPLE_DICT = {
    "foo": "bar",
    "answer": 42,
    "hello world": [1, 2, 3.14],
    "empty": None,
}
SAMPLE_DICT_STR = json.dumps(SAMPLE_DICT)

SAMPLE_ARRAY = ["foo", 1, [3.14], {"hello": "world"}, None]
SAMPLE_ARRAY_STR = json.dumps(SAMPLE_ARRAY)


def test_llm_parse_json_dict():
    assert llm_parse_json(SAMPLE_DICT_STR) == SAMPLE_DICT


def test_llm_parse_json_array():
    assert llm_parse_json(SAMPLE_ARRAY_STR) == SAMPLE_ARRAY


def test_llm_parse_json_handle_single_quotes():
    assert llm_parse_json(SAMPLE_DICT_STR.replace('"', "'")) == SAMPLE_DICT
    assert llm_parse_json(SAMPLE_ARRAY_STR.replace('"', "'")) == SAMPLE_ARRAY


def test_llm_parse_json_none_string():
    # Handles edge case where LLM produces "None" instead of "null"
    assert llm_parse_json(SAMPLE_DICT_STR.replace("null", "None")) == SAMPLE_DICT
    assert llm_parse_json(SAMPLE_ARRAY_STR.replace("null", "None")) == SAMPLE_ARRAY


def test_llm_parse_json_invalid():
    # Raises error for invalid string
    with pytest.raises(ValueError, match="Failed to parse JSON"):
        llm_parse_json("[invalid]")



import click
import uvicorn
from pydantic import create_model
from pydantic.fields import FieldInfo

from otaro.api.server import app
from otaro.task import Task


@click.command()
@click.argument("config_file")
def deploy(config_file: str):
    print(config_file)
    task = Task.from_config(config_file)
    input_field_attributes = {}
    for field in task.inputs:
        input_field_attributes[field.name] = (field.model, FieldInfo())
    PayloadClass = create_model(
        "Payload",
        **input_field_attributes,
    )

    async def run(payload: PayloadClass):
        return await task.run(optimize=False, **payload.model_dump(mode="json"))

    app.add_api_route("/run", endpoint=run, methods=["POST"])
    uvicorn.run(app)



import otaro.rules as rules

TEST_STRING = "foobar"
TEST_INT = 1


def test_eq():
    assert rules.eq(TEST_INT, TEST_INT) is True
    assert rules.eq(TEST_STRING, TEST_STRING) is True

    assert rules.eq(TEST_INT, TEST_STRING) is False


def test_neq():
    assert rules.neq(TEST_INT, TEST_STRING) is True

    assert rules.neq(TEST_INT, TEST_INT) is False
    assert rules.neq(TEST_STRING, TEST_STRING) is False


def test_contains():
    assert rules.contains([TEST_INT], TEST_INT) is True
    assert rules.contains(TEST_STRING, TEST_STRING[:1]) is True

    assert rules.contains([TEST_INT], TEST_STRING) is False
    assert rules.contains(TEST_STRING, TEST_STRING[::-1]) is False


def test_length_eq():
    assert rules.length_eq([TEST_INT], len([TEST_INT])) is True
    assert rules.length_eq(TEST_STRING, len(TEST_STRING)) is True



import importlib
import inspect
import re

from pydantic import BaseModel
from pyparsing import (
    Literal,
    QuotedString,
    Word,
    delimited_list,
    identbodychars,
    identchars,
    nums,
    one_of,
)

string_arg = QuotedString("'", esc_char="\\", unquote_results=False) | QuotedString(
    '"', esc_char="\\", unquote_results=False
)

number_arg = Word(nums) | Word(nums) + "." + Word(nums)

boolean_arg = Literal("True") | Literal("False")

array_item = string_arg | number_arg
array_list = delimited_list(array_item, allow_trailing_delim=True)
array_arg = Literal("[") + array_list + Literal("]")
tuple_arg = Literal("(") + array_list + Literal(")")

arg_name = Word(identchars, identbodychars)
arg_value = string_arg | number_arg | boolean_arg | tuple_arg | array_arg
generic_arg_value = string_arg | number_arg | boolean_arg | tuple_arg | array_arg
arg_item = (arg_name + Literal("=").suppress() + arg_value) | arg_value
arg_list = delimited_list(arg_item)


def get_value_parser(sample: BaseModel | None = None):
    if sample is None:
        sample_attrs = []
    else:
        sample_attrs = list(
            sample.model_dump(mode="json").keys()
        )  # Note this does not support nested keys

    def value_parser(string, location, tokens):
        key = None
        emit_tokens = []
        if tokens[0] == "[":
            # Arrays
            emit_tokens = eval("[" + ",".join(tokens[1:-1]) + "]")
        elif tokens[0] == "(":
            # Tuples
            emit_tokens = eval("(" + ",".join(tokens[1:-1]) + ")")
        elif tokens[0].startswith('"') or tokens[0].startswith("'"):
            # Quoted strings
            emit_tokens = eval(tokens[0])
        elif tokens[0] in sample_attrs:
            # Sample attributes
            emit_tokens = sample.model_dump(mode="json")[tokens[0]]
        elif len(tokens) > 2 and tokens[1] == "=":
            # Kwarg
            key = tokens[0]
            emit_tokens = value_parser(None, None, tokens[2:])[1]
        else:
            # Everything else
            emit_tokens = eval(tokens[0])

        return [key, emit_tokens]

    return value_parser


# TODO: Supported nested inputs / outputs
def process_signature(args_str: str, sample: BaseModel | None = None):
    if sample is None:
        sample_attrs = []
    else:
        sample_attrs = list(
            sample.model_dump(mode="json").keys()
        )  # Note this does not support nested keys
    field_arg = one_of(sample_attrs)
    arg_value = generic_arg_value | field_arg
    arg_item = (arg_name + Literal("=") + arg_value) | arg_value
    arg_item.set_parse_action(get_value_parser(sample))
    arg_list = delimited_list(arg_item)
    arr = arg_list.parse_string(args_str, parse_all=True)
    args = []
    kwargs = {}
    for i in range(0, len(arr), 2):
        key = arr[i]
        value = arr[i + 1]
        if key is None:
            args.append(value)
        else:
            kwargs[key] = value
    return args, kwargs


# TODO
# def validate_rule_str(rule_str: str, input_fields: list, output_fields: list):
#     raise NotImplementedError


def eval_rule_str(rule_str: str, sample: BaseModel | None = None):
    fn_pattern = re.compile(
        "^(?P<module>.*?)\\.(?P<fn>[a-zA-Z_]*?)(\\((?P<args>.*)\\))?$", re.DOTALL
    )
    match = fn_pattern.match(rule_str.strip())
    if match:
        module_name = match.groupdict().get("module")
        fn_name = match.groupdict().get("fn")
        args_str = match.groupdict().get("args")
        module = importlib.import_module(module_name)
        fn = getattr(module, fn_name)
        if args_str:
            args, kwargs = process_signature(args_str, sample)
            return fn(*args, **kwargs)
        return fn(sample)
    raise ValueError(f"Invalid rule - {rule_str}")


def get_rule_source(rule_str: str):
    fn_pattern = re.compile(
        "^(?P<module>.*?)\\.(?P<fn>[a-zA-Z_]*?)(\\((?P<args>.*)\\))?$", re.DOTALL
    )
    match = fn_pattern.match(rule_str.strip())
    if match:
        module_name = match.groupdict().get("module")
        fn_name = match.groupdict().get("fn")
        module = importlib.import_module(module_name)
        fn = getattr(module, fn_name)
        return f"Rule: {rule_str}\n{inspect.getsource(fn)}"

    raise ValueError(f"Invalid rule - {rule_str}")
